########################################################################################
#    Peelpreter is a interpreter designed to interpret the language, Monkey.
#    Copyright (C) 2024 Jeebak Samajdwar
#
#    This file is part of Peelpreter
#
#    Peelpreter is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    any later version.
#
#    Peelpreter is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################################

from sys import setrecursionlimit
from typing import Union

from .. import astt
from ..objectt.enviroment import Enviroment
from .. import error
from ..evaluator.mbuiltins import builtins
from .. import objectt as obj

setrecursionlimit(10**6)


def evaluate(
    node: astt.Node, env: Enviroment, fname="stdin"
) -> obj.Object:
    def istruthy(cond_object):
        if cond_object == obj.NULL:
            return False
        elif cond_object == obj.TRUE:
            return True
        elif cond_object == obj.FALSE:
            return False
        else:
            return True

    def is_error(tobject):
        if tobject is not None:
            return tobject.type() == obj.OBJ_ERROR
        return False

    def eval_program(statements):
        result = obj.Object()
        for statement in statements:
            result = evaluate(statement, env)

            if type(result) == obj.ReturnValue:
                return result.value
            elif type(result) == obj.Error:
                return result

        return result

    def eval_blockstmt(block: astt.BlockStatement):
        result: Union[obj.Object, None] = obj.Object()

        for statement in block.statements:
            result = evaluate(statement, env)

            if result is not None:
                if (
                    result.type() == obj.OBJ_RETURN_VALUE
                    or result.type() == obj.OBJ_ERROR
                ):
                    return result

        return result

    def eval_prefixexpr(operator, right):
        match operator:
            case "!":
                return eval_bang(right)
            case "-":
                return eval_minus(right)
            case _:
                return obj.Error(error.
                    UnknownOperator(fname, operator, None, right.type(), (-1, -1))
                )

    def eval_infixexpr(operator, left, right):
        if left.type() == obj.OBJ_NUM and right.type() == obj.OBJ_NUM:
            return eval_num_infixexpr(operator, left, right)
        elif left.type() == obj.OBJ_STRING and right.type() == obj.OBJ_STRING:
            return eval_str_infixexpr(operator, left, right)
        elif operator == "==":
            return obj.TRUE if left == right else obj.FALSE
        elif operator == "!=":
            return obj.TRUE if left != right else obj.FALSE
        else:
            return obj.Error(error.
                UnknownOperator(fname, operator, left.type(), right.type(), (-1, -1))
            )

    def eval_num_infixexpr(operator, left, right):
        leftval = left.value
        rightval = right.value

        match operator:
            case "+":
                return obj.Number(leftval + rightval)
            case "-":
                return obj.Number(leftval - rightval)
            case "*":
                return obj.Number(leftval * rightval)
            case "/":
                return obj.Number(leftval / rightval)
            case "<":
                return obj.TRUE if leftval < rightval else obj.FALSE
            case ">":
                return obj.TRUE if leftval > rightval else obj.FALSE
            case "==":
                return obj.TRUE if leftval == rightval else obj.FALSE
            case "!=":
                return obj.TRUE if leftval != rightval else obj.FALSE
            case _:
                return obj.Error(error.
                    UnknownOperator(
                        fname, operator, left.type(), right.type(), (-1, -1)
                    )
                )

    def eval_str_infixexpr(operator, left, right):
        if operator != "+":
            return obj.Error(error.
                UnknownOperator(fname, operator, left.type(), right.type(), (-1, -1))
            )
        left_val = left.value
        right_val = right.value

        return obj.String(left_val + right_val)

    def eval_exprs(expressions, env):
        result = []
        for expression in expressions:
            evaluated = evaluate(expression, env)
            if is_error(evaluated):
                return [evaluated]
            result.append(evaluated)

        return result

    def eval_ifexpr(ifelse):
        condition = evaluate(ifelse.condition, env)
        if is_error(condition):
            return condition

        if istruthy(condition):
            return evaluate(ifelse.consequence, env)
        elif ifelse.alternative.literal != "":
            return evaluate(ifelse.alternative, env)
        else:
            return obj.NULL

    def eval_hashlit(node, env):
        pairs = dict()

        for key_node, val_node in node.pairs.items():
            key = evaluate(key_node, env)
            if is_error(key):
                return key
            if not isinstance(key, obj.Hashable):
                return obj.Error(error.UnsupporteKeyType(fname, key.type(), (-1, -1)))
 
            value = evaluate(val_node, env)
            if is_error(value):
                return value
            hashed = key.hash_key()
            pairs[hashed] = obj.HashPair(key, value)

        return obj.Hash(pairs)

    def eval_indexexpr(left, indexexpr):
        if left.type() == obj.OBJ_ARRAY and indexexpr.type() == obj.OBJ_NUM:
            return eval_arr_indexexpr(left, indexexpr)
        elif left.type() == obj.OBJ_HASH:
            return eval_hash_indexexpr(left, indexexpr)
        elif indexexpr.type() != obj.OBJ_NUM:
            return obj.Error(error.UnsupportedIndexType(fname, indexexpr.type(), (-1, -1)))
        else:
            return obj.Error(error.UnsupportedIndexAccessType(fname, left.type(), (-1, -1)))

    def eval_hash_indexexpr(hash, indexexpr):
        if not isinstance(indexexpr, obj.Hashable):
            return obj.Error(error.UnsupporteKeyType(fname, indexexpr.type(), (-1, -1)))
        pair = hash.pairs.get(indexexpr.hash_key())
        if pair is None:
            return obj.NULL
        return pair.value

    def eval_hash_reassign(hash, key, value):
        if not isinstance(key, obj.Hashable):
            return obj.Error(error.UnsupporteKeyType(fname, key.type(), (-1, -1)))

        hash.pairs[key.hash_key()] = obj.HashPair(key, value)

        return value

    def eval_arr_indexexpr(array, indexexpr):
        index = int(indexexpr.value)
        maximum = len(array.elements) - 1

        if index < 0 or index > maximum:
            return obj.NULL
        return array.elements[index]

    def eval_arr_reassign(array, indexexpr, value):
        index = int(indexexpr.value)
        maximum = len(array.elements) - 1

        if index < 0 or index > maximum:
            return obj.Error("Out of Bounds")

        array.elements[index] = value

        return value

    def eval_identifier(node, env: Enviroment):
        value = env.get(node.literal)
        if value is not None:
            return value
        builtin = builtins.get(node.literal)
        if builtin is not None:
            return builtin

        return obj.Error(error.UnknownIdentifier(fname, node.literal, (-1, -1)))

    def eval_minus(right):
        if type(right) != obj.Number:
            return obj.Error(error.UnknownOperator(fname, "-", None, right.type(), (-1, -1)))
        value = right.value
        return obj.Number(-value)

    def eval_bang(right):
        if right == obj.TRUE:
            return obj.FALSE
        elif right == obj.FALSE:
            return obj.TRUE
        elif right == obj.NULL:
            return obj.TRUE
        else:
            return obj.FALSE

    def apply_func(func, arguments):
        if type(func) == obj.Function:
            extended_env = extend_funcenv(func, arguments)
            evaluated = evaluate(func.body, extended_env)
            return unwrap_rtrvalue(evaluated)
        elif type(func) == obj.Builtin:
            return func.func(fname, arguments)

        return obj.Error(error.NotAFunction(fname, func, (-1, -1)))

    def extend_funcenv(func: obj.Function, arguments):
        env = Enviroment(func.env)
        for index, parameter in enumerate(func.parametres):
            env.set_iden(parameter.literal, arguments[index])
        return env

    def unwrap_rtrvalue(tobject):
        if type(tobject) == obj.ReturnValue:
            return tobject.value
        return tobject

    if type(node) == astt.Program:
        return eval_program(node.statements)
    elif type(node) == astt.ExpressionStatement:
        return evaluate(node.expression, env)
    elif type(node) == astt.PrefixExpression:
        rightexpr = evaluate(node.rightexpr, env)
        if is_error(rightexpr):
            return rightexpr
        return eval_prefixexpr(node.operator, rightexpr)
    elif type(node) == astt.InfixExpression:
        left = evaluate(node.leftexpr, env)
        if is_error(left):
            return left
        right = evaluate(node.rightexpr, env)
        if is_error(right):
            return right
        return eval_infixexpr(node.operator, left, right)
    elif type(node) == astt.BlockStatement:
        return eval_blockstmt(node)
    elif type(node) == astt.FunctionLiteral:
        parametres = node.parameters
        body = node.body
        return obj.Function(parametres, body, env)
    elif type(node) == astt.CallExpression:
        function = evaluate(node.function, env)
        if is_error(function):
            return function
        arguments = eval_exprs(node.arguments, env)
        if len(arguments) == 1 and is_error(arguments[0]):
            return arguments[0]
        return apply_func(function, arguments)
    elif type(node) == astt.HashLiteral:
        return eval_hashlit(node, env)
    elif type(node) == astt.ArrayLiteral:
        elements = eval_exprs(node.elements, env)
        if len(elements) == 1 and is_error(elements[0]):
            return elements[0]
        return obj.Array(elements)
    elif type(node) == astt.IndexExpression:
        leftexpr = evaluate(node.left, env)
        if is_error(leftexpr):
            return leftexpr
        indexexpr = evaluate(node.index, env)
        if is_error(indexexpr):
            return indexexpr
        return eval_indexexpr(leftexpr, indexexpr)
    elif type(node) == astt.IfExpression:
        return eval_ifexpr(node)
    elif type(node) == astt.ReturnStatement:
        value = evaluate(node.valuexp, env)
        if is_error(value):
            return value
        return obj.ReturnValue(value)
    elif type(node) == astt.LetStatement:
        value = evaluate(node.value, env)
        if is_error(value):
            return value
        env.set_iden(name=node.name.literal, value=value)
        return value
    elif type(node) == astt.ReassignmentStatement:
        structure = evaluate(node.index_expr.left, env)
        index = evaluate(node.index_expr.index, env)
        value = evaluate(node.value, env)
        if isinstance(structure, obj.Array):
            return eval_arr_reassign(structure, index, value)
        elif isinstance(structure, obj.Hash):
            return eval_hash_reassign(structure, index, value)
        else:
            return obj.Error("Unsupported")

    elif type(node) == astt.Identifier:
        return eval_identifier(node, env)
    elif type(node) == astt.Number:
        return obj.Number(node.value)
    elif type(node) == astt.String:
        return obj.String(node.string)
    elif type(node) == astt.Boolean:
        if node.value:
            return obj.TRUE
        return obj.FALSE
    elif type(node) == astt.Null:
        return obj.NULL

    return obj.Error(error.UnknownNode(fname, (-1, -1)))
